import torch
import torch.nn.functional as F
from torch import nn

from helios import nn as hln


@hln.NETWORK_REGISTRY.register
class ClassifierNet(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)  # flatten all dimensions except batch
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class TestNewtorks:
    def test_registry(self, check_registry) -> None:
        check_registry(hln.NETWORK_REGISTRY, ["ClassifierNet"])

    def test_create(self, check_create_function) -> None:
        check_create_function(hln.NETWORK_REGISTRY, hln.create_network)

    def test_ema(self) -> None:
        net = nn.Conv2d(1, 20, 5)
        ema = hln.swa_utils.EMA(net)
        ema.update(net)
        n = ema.module
        assert net != n



import pathlib
import tomllib


def generate_requirements_file(
    repo_dir: pathlib.Path, name: str, req_list: list[str], header: list[str]
) -> None:
    """
    Generate a requirements file located under <repo_dir>/requirements/<name>.txt.

    Args:
        repo_dir (Path): the root of the repository.
        name (str): the name of the file to save.
        req_list (list[str]): the list of requirements.
        header (list[str]): the header for the files (may be empty).
    """
    req_list = sorted(req_list)
    req_root = repo_dir / "requirements"
    req_root.mkdir(exist_ok=True, parents=True)
    req_fname = req_root / f"{name}.txt"
    req_fname.write_text(
        "\n".join(header + req_list) + "\n", encoding="utf-8", newline="\n"
    )


def main() -> None:
    """Generate the requirements files for each dependency list in pyproject.toml."""
    script_path = pathlib.Path(__file__)
    repo_dir = script_path.parent.parent
    script_relative = script_path.relative_to(repo_dir)
    header = [
        f"# Generated via {str(script_relative.as_posix())} and pre-commit hook.",
        "# Do not edit this file; modify pyproject.toml instead.",
    ]

    pyproject = tomllib.loads((repo_dir / "pyproject.toml").read_text(encoding="utf-8"))

    generate_requirements_file(
        repo_dir, "default", pyproject["project"]["dependencies"], header
    )

    default_deps = pyproject["project"]["dependencies"]

    for key, opt_list in pyproject["project"]["optional-dependencies"].items():
        generate_requirements_file(repo_dir, key, default_deps + opt_list, header)


if __name__ == "__main__":
    main()



import pathlib
import typing

import optuna
import pytest
import torch

import helios.model as hlm
import helios.plugins as hlp
import helios.trainer as hlt
from helios.plugins.optuna import _ORIG_NUMBER_KEY, OptunaPlugin


class ExceptionPlugin(hlp.Plugin):
    def __init__(self, exc_type: type[Exception] | list[type[Exception]]):
        super().__init__(f"exception_{str(exc_type)}")
        self._exc_type = exc_type

    def setup(self) -> None:
        pass

    def configure_trainer(self, t: hlt.Trainer) -> None:
        self._append_train_exceptions(self._exc_type, t)
        self._append_test_exceptions(self._exc_type, t)


class PluginModel(hlm.Model):
    def __init__(self, save_name: str):
        super().__init__(save_name)

    def setup(self, fast_init: bool = False) -> None:
        pass


class TestPlugins:
    def test_registry(self, check_registry) -> None:
        check_registry(hlp.PLUGIN_REGISTRY, ["CUDAPlugin", "OptunaPlugin"])

    def test_create(self, check_create_function) -> None:
        check_create_function(hlp.PLUGIN_REGISTRY, hlp.create_plugin)

    def test_append_exceptions(self) -> None:
        t = hlt.Trainer()
        plugin = ExceptionPlugin(RuntimeError)
        exc_list: list[type[Exception]] = [RuntimeError]

        plugin.configure_trainer(t)
        assert t.train_exceptions == exc_list
        assert t.test_exceptions == exc_list

        exc_list.extend([ValueError, TypeError])
        plugin = ExceptionPlugin([ValueError, TypeError])
        plugin.configure_trainer(t)

        assert t.train_exceptions == exc_list
        assert t.test_exceptions == exc_list


class TestCUDAPlugin:
    def check_batch_device(self, x: typing.Any, device: torch.device) -> None:
        if isinstance(x, torch.Tensor):
            assert x.device == device
        elif isinstance(x, dict):
            assert all(val.device == device for _, val in x.items())
        else:
            assert all(elem.device == device for elem in x)

    def check_batch_processing(
        self, plugin: hlp.CUDAPlugin, x: typing.Any, device: torch.device
    ) -> None:
        ret = plugin.process_training_batch(x, hlt.TrainingState())
        self.check_batch_device(ret, device)

        ret = plugin.process_validation_batch(x, 0)
        self.check_batch_device(ret, device)

        ret = plugin.process_testing_batch(x, 0)
        self.check_batch_device(ret, device)

    def test_plugin_id(self) -> None:
        assert hasattr(hlp.CUDAPlugin, "plugin_id")
        assert hlp.CUDAPlugin.plugin_id == "cuda"

    def test_batch_processing(self) -> None:
        if not torch.cuda.is_available():
            return

        def create_tensor() -> torch.Tensor:
            return torch.randn((1, 3, 32, 32)).to("cpu")

        device = torch.device("cuda:0")

        plugin = hlp.CUDAPlugin()
        plugin.is_distributed = False
        plugin.map_loc = {"cuda:0": "cuda:0"}
        plugin.device = device

        assert plugin.unique_overrides.training_batch
        assert plugin.unique_overrides.validation_batch
        assert plugin.unique_overrides.testing_batch

        self.check_batch_processing(plugin, create_tensor(), device)
        self.check_batch_processing(plugin, [create_tensor(), create_tensor()], device)
        self.check_batch_processing(plugin, (create_tensor(), create_tensor()), device)
        self.check_batch_processing(
            plugin, {"a": create_tensor(), "b": create_tensor()}, device
        )

    def test_configure(self) -> None:
        if not torch.cuda.is_available():
            return

        trainer = hlt.Trainer()
        plugin = hlp.CUDAPlugin()
        plugin.configure_trainer(trainer)

        assert "cuda" in trainer.plugins
        assert trainer.plugins["cuda"] == plugin


# Ignore the warnings coming from optuna.
@pytest.mark.filterwarnings(
    ("ignore::optuna.exceptions.ExperimentalWarning"), ("ignore::FutureWarning")
)
class TestOptunaPlugin:
    def test_plugin_id(self) -> None:
        assert hasattr(OptunaPlugin, "plugin_id")
        assert OptunaPlugin.plugin_id == "optuna"

    def test_invalid_storage(self) -> None:
        def objective(trial: optuna.Trial) -> int:
            plugin = OptunaPlugin(trial, "accuracy")
            plugin.is_distributed = True
            with pytest.raises(ValueError):
                plugin.setup()

            return 0

        study = optuna.create_study()
        study.optimize(objective, n_trials=1)

    def test_configure(self) -> None:
        def objective(trial: optuna.Trial) -> int:
            plugin = OptunaPlugin(trial, "accuracy")
            trainer = hlt.Trainer()

            plugin.configure_trainer(trainer)
            assert len(trainer.plugins) == 1
            assert trainer.plugins["optuna"] == plugin
            assert len(trainer.train_exceptions) == 1
            assert trainer.train_exceptions[0] == optuna.TrialPruned

            model = PluginModel("plugin-model")
            plugin.configure_model(model)
            assert model.save_name == "plugin-model_trial-0"

            return 0

        study = optuna.create_study()
        study.optimize(objective, n_trials=1)

    def check_in_range(
        self, val: typing.Any, t: type, low: typing.Any, high: typing.Any
    ) -> None:
        assert isinstance(val, t)
        assert low <= val <= high

    def check_in_sequence(self, val: typing.Any, t: type, seq: typing.Container) -> None:
        assert isinstance(val, t)
        assert val in seq

    def test_suggest(self) -> None:
        def objective(trial: optuna.Trial) -> int:
            plugin = OptunaPlugin(trial, "accuracy")

            with pytest.raises(KeyError):
                plugin.suggest("foo", "bar")

            seq = [1, 2, 3]
            self.check_in_sequence(
                plugin.suggest("categorical", "val1", choices=seq), int, seq
            )

            low = 0
            high = 10
            self.check_in_range(
                plugin.suggest("int", "val2", low=low, high=high), int, low, high
            )

            high = 1
            self.check_in_range(
                plugin.suggest("float", "val3", low=low, high=high), float, low, high
            )

            return 0

        study = optuna.create_study()
        study.optimize(objective, n_trials=1)

    def test_state_dict(self) -> None:
        def objective(trial: optuna.Trial) -> int:
            plugin = OptunaPlugin(trial, "accuracy")
            x = plugin.suggest("float", "x", low=-10, high=10)

            state_dict = plugin.state_dict()
            assert len(state_dict) == 1
            assert "x" in state_dict
            assert state_dict["x"] == x

            return 0

        study = optuna.create_study()
        study.optimize(objective, n_trials=1)

    def test_resume_trial(self, tmp_path: pathlib.Path) -> None:
        num_trials = 10
        successful_trials = [False for _ in range(num_trials)]
        offset = 0

        def objective(trial: optuna.Trial) -> float:
            nonlocal offset
            plugin = OptunaPlugin(trial, "accuracy")
            plugin.setup()

            model = PluginModel("plugin-model")
            plugin.configure_model(model)

            trial_num = trial.number
            if _ORIG_NUMBER_KEY in trial.user_attrs:
                trial_num = trial.user_attrs[_ORIG_NUMBER_KEY]
                assert model.save_name == f"plugin-model_trial-{trial_num}"

                # Artificially offset the number by 1 so that when we subtract 1 later on
                # the indices work out correctly.
                trial_num += 1
                offset = 1

            if trial.number == num_trials / 2:
                raise RuntimeError("half-way stop")

            successful_trials[trial_num - offset] = True

            return 0

        def create_study() -> optuna.Study:
            storage_path = tmp_path / "trial_test.db"
            return optuna.create_study(
                study_name="trial_test",
                storage=f"sqlite:///{storage_path}",
                load_if_exists=True,
            )

        def optimize(study: optuna.Study) -> None:
            study.optimize(
                objective,
                n_trials=num_trials,
                callbacks=[
                    optuna.study.MaxTrialsCallback(
                        num_trials,
                        states=(optuna.trial.TrialState.COMPLETE,),
                    )
                ],
            )

        study = create_study()
        with pytest.raises(RuntimeError):
            optimize(study)

        del study

        study = create_study()

        OptunaPlugin.enqueue_failed_trials(study)
        optimize(study)

        for v in successful_trials:
            assert v



import pathlib
import typing

import cv2
import numpy as np
import numpy.typing as npt
import PIL
import torch
from torch.utils import data as tud

from helios import core, data
from helios.core import rng
from helios.data import functional as hldf
from helios.data import samplers as hlds
from helios.data import transforms as hldt

DATASET_SIZE: int = 16


class RandomDataset(tud.Dataset):
    def __getitem__(self, index):
        gen = rng.get_default_numpy_rng().generator
        return gen.integers(0, 1000, 3)

    def __len__(self):
        return DATASET_SIZE


class SequentialDataset(tud.Dataset):
    def __init__(self):
        super().__init__()

        self._samples = [
            torch.tensor([359, 639, 636]),
            torch.tensor([963, 738, 996]),
            torch.tensor([766, 97, 292]),
            torch.tensor([500, 324, 687]),
            torch.tensor([15, 841, 180]),
            torch.tensor([553, 266, 928]),
            torch.tensor([269, 806, 472]),
            torch.tensor([694, 632, 678]),
            torch.tensor([92, 228, 424]),
            torch.tensor([421, 568, 705]),
            torch.tensor([686, 769, 903]),
            torch.tensor([440, 610, 126]),
            torch.tensor([809, 833, 629]),
            torch.tensor([340, 810, 512]),
            torch.tensor([23, 322, 577]),
            torch.tensor([757, 651, 429]),
        ]

    def __getitem__(self, index):
        return self._samples[index]

    def __len__(self):
        return DATASET_SIZE


class SampleDataModule(data.DataModule):
    def setup(self) -> None:
        params = data.DataLoaderParams(
            batch_size=2, num_workers=2, random_seed=rng.get_default_seed()
        )

        params.shuffle = True
        self._train_dataset = self._create_dataset(RandomDataset(), params)

        params.num_workers = 2
        self._valid_dataset = self._create_dataset(SequentialDataset(), params)

        params.shuffle = False
        self._test_dataset = self._create_dataset(SequentialDataset(), params)


class TestTransforms:
    def check_type(self, x, exp_type: type) -> None:
        assert isinstance(x, exp_type)

        if isinstance(exp_type, list | tuple):
            for elem in x:  # type: ignore[attr-defined]
                assert isinstance(elem, torch.Tensor)

    def check_to_tensor(self, img, exp_type: type, exp_shape: tuple[int, ...]) -> None:
        to_tensor = hldt.create_transform("ToImageTensor")
        as_tens: torch.Tensor | list[torch.Tensor] | tuple[torch.Tensor, ...] = to_tensor(
            img
        )

        self.check_type(as_tens, exp_type)

        if isinstance(as_tens, list | tuple):
            for t in as_tens:
                assert t.shape == exp_shape
        else:
            assert as_tens.shape == exp_shape

    def generate_sample(
        self,
        size: tuple[int, ...],
        num_elems: int,
        as_tuple: bool = False,
        as_pil: bool = False,
    ):
        gen = rng.get_default_numpy_rng().generator

        samples: list[npt.NDArray | PIL.Image.Image] = []
        for _ in range(num_elems):
            base = np.uint8(gen.random(size=size)) * 255
            if as_pil:
                samples.append(PIL.Image.fromarray(base))
            else:
                samples.append(base)

        if as_tuple:
            return tuple(samples)
        return samples

    def test_to_tensor(self) -> None:
        assert "ToImageTensor" in hldt.TRANSFORM_REGISTRY

        rng.seed_rngs()
        size = (32, 32, 3)
        exp_size = (3, 32, 32)

        # Single image (numpy)
        self.check_to_tensor(self.generate_sample(size, 1)[0], torch.Tensor, exp_size)

        # Multi-image list (numpy)
        self.check_to_tensor(self.generate_sample(size, 2), list, exp_size)

        # Multi-image tuple (numpy)
        self.check_to_tensor(
            self.generate_sample(size, 2, as_tuple=True),
            tuple,
            exp_size,
        )

        # Single image (PIL)
        self.check_to_tensor(
            self.generate_sample(size, 1, as_pil=True)[0], torch.Tensor, exp_size
        )

        # Multi-image list (PIL)
        self.check_to_tensor(self.generate_sample(size, 2, as_pil=True), list, exp_size)

        # Multi-image tuple (PIL)
        self.check_to_tensor(
            self.generate_sample(size, 2, as_tuple=True, as_pil=True),
            tuple,
            exp_size,
        )


class TestFunctional:
    def check_image(self, gt: npt.NDArray, out_path: pathlib.Path, **kwargs) -> None:
        # Ensure that we convert to "BGR" so the output comes out the way we expect it to
        # come out.
        if kwargs.get("as_rgb", True):
            if len(gt.shape) == 2:
                tmp = gt
            elif len(gt.shape) == 3 and gt.shape[2] == 3:
                tmp = cv2.cvtColor(gt, cv2.COLOR_RGB2BGR)
            elif len(gt.shape) == 3 and gt.shape[2] == 4:
                tmp = cv2.cvtColor(gt, cv2.COLOR_RGBA2BGRA)
        else:
            tmp = gt

        cv2.imwrite(str(out_path), tmp)

        ret = hldf.load_image(out_path, **kwargs)
        assert isinstance(ret, np.ndarray)
        assert np.array_equal(gt, ret)

    def test_load_image(self, tmp_path: pathlib.Path):
        rng.seed_rngs()
        gen = rng.get_default_numpy_rng().generator
        out_path = tmp_path / "tmp.png"
        max_8 = 255
        max_16 = 65535

        # 8-bit grayscale.
        self.check_image(
            (gen.random(size=(32, 32)) * max_8).astype(np.uint8),
            out_path,
            flags=cv2.IMREAD_GRAYSCALE,
        )

        # 16-bit grayscale
        self.check_image(
            (gen.random(size=(32, 32)) * max_16).astype(np.uint16),
            out_path,
            flags=cv2.IMREAD_UNCHANGED,
        )

        # 8-bit RGB
        self.check_image(
            (gen.random(size=(32, 32, 3)) * max_8).astype(np.uint8),
            out_path,
        )

        # 16-bit RGB
        self.check_image(
            (gen.random(size=(32, 32, 3)) * max_16).astype(np.uint16),
            out_path,
            flags=cv2.IMREAD_UNCHANGED,
        )

        # 8-bit RGBA
        self.check_image(
            (gen.random(size=(32, 32, 4)) * max_8).astype(np.uint8),
            out_path,
            flags=cv2.IMREAD_UNCHANGED,
        )

        # 16-bit RGB
        self.check_image(
            (gen.random(size=(32, 32, 4)) * max_16).astype(np.uint16),
            out_path,
            flags=cv2.IMREAD_UNCHANGED,
        )

        # Load image as is.
        self.check_image(
            (gen.random(size=(32, 32, 4)) * max_8).astype(np.uint8),
            out_path,
            flags=cv2.IMREAD_UNCHANGED,
            as_rgb=False,
        )

    def test_load_image_pil(self, tmp_path: pathlib.Path):
        rng.seed_rngs()
        gen = rng.get_default_numpy_rng().generator
        img = (gen.random(size=(32, 32, 3)) * 255).astype(np.uint8)
        as_pil = PIL.Image.fromarray(img)

        out_path = tmp_path / "tmp.png"
        cv2.imwrite(str(out_path), cv2.cvtColor(img, cv2.COLOR_BGR2RGB))

        ret = hldf.load_image_pil(out_path)
        assert isinstance(ret, np.ndarray)
        assert np.array_equal(img, ret)

        ret = hldf.load_image_pil(out_path, as_numpy=False)
        assert isinstance(ret, PIL.Image.Image)
        diff = PIL.ImageChops.difference(as_pil, ret)  # type: ignore[attr-defined]
        assert diff.getbbox() is None

    def test_tensor_to_numpy(self):
        tens = torch.randn((3, 32, 32))
        ret = hldf.tensor_to_numpy(tens)
        assert isinstance(ret, np.ndarray)
        assert ret.shape == (32, 32, 3)


class TestDataModule:
    def get_expected_train_batches(self) -> list[torch.Tensor]:
        return [
            torch.tensor([[359, 639, 636], [963, 738, 996]]),
            torch.tensor([[766, 97, 292], [500, 324, 687]]),
            torch.tensor([[15, 841, 180], [553, 266, 928]]),
            torch.tensor([[269, 806, 472], [694, 632, 678]]),
            torch.tensor([[92, 228, 424], [421, 568, 705]]),
            torch.tensor([[686, 769, 903], [440, 610, 126]]),
            torch.tensor([[809, 833, 629], [340, 810, 512]]),
            torch.tensor([[23, 322, 577], [757, 651, 429]]),
        ]

    def get_expected_valid_batches(self) -> list[torch.Tensor]:
        return [
            torch.tensor([[809, 833, 629], [694, 632, 678]]),
            torch.tensor([[359, 639, 636], [553, 266, 928]]),
            torch.tensor([[500, 324, 687], [963, 738, 996]]),
            torch.tensor([[421, 568, 705], [15, 841, 180]]),
            torch.tensor([[766, 97, 292], [23, 322, 577]]),
            torch.tensor([[269, 806, 472], [757, 651, 429]]),
            torch.tensor([[340, 810, 512], [440, 610, 126]]),
            torch.tensor([[92, 228, 424], [686, 769, 903]]),
        ]

    def get_expected_test_batches(self) -> list[torch.Tensor]:
        return [
            torch.tensor([[359, 639, 636], [963, 738, 996]]),
            torch.tensor([[766, 97, 292], [500, 324, 687]]),
            torch.tensor([[15, 841, 180], [553, 266, 928]]),
            torch.tensor([[269, 806, 472], [694, 632, 678]]),
            torch.tensor([[92, 228, 424], [421, 568, 705]]),
            torch.tensor([[686, 769, 903], [440, 610, 126]]),
            torch.tensor([[809, 833, 629], [340, 810, 512]]),
            torch.tensor([[23, 322, 577], [757, 651, 429]]),
        ]

    def check_batches(self, exp: list[torch.Tensor], val: list[torch.Tensor]) -> None:
        for b, exp_b in zip(exp, val, strict=True):
            assert torch.all(b == exp_b)

    def prepare(
        self, split: data.DatasetSplit, skip_seed: bool = False
    ) -> tuple[tud.DataLoader, tud.Sampler]:
        if not skip_seed:
            rng.seed_rngs()

        datamodule = SampleDataModule()
        datamodule.setup()

        if split == data.DatasetSplit.TRAIN:
            return core.get_from_optional(datamodule.train_dataloader())
        if split == data.DatasetSplit.VALID:
            return core.get_from_optional(datamodule.valid_dataloader())
        return core.get_from_optional(datamodule.test_dataloader())

    def test_worker_seeding(self) -> None:
        dataloader, _ = self.prepare(data.DatasetSplit.TRAIN)

        batches = []
        for batch in dataloader:
            batches.append(batch)

        exp_batches = self.get_expected_train_batches()
        self.check_batches(exp_batches, batches)

    def test_resume_random(self) -> None:
        dataloader, sampler = self.prepare(data.DatasetSplit.VALID)
        assert isinstance(sampler, hlds.ResumableRandomSampler)

        half_step = len(dataloader) // 2
        exp_batches = self.get_expected_valid_batches()

        batches = []
        sampler.set_epoch(0)
        for step, batch in enumerate(dataloader):
            if step >= half_step:
                break
            batches.append(batch)

        self.check_batches(exp_batches[:half_step], batches)

        # Grab the state and clear everything.
        rng_state = rng.get_rng_state_dict()

        del dataloader, sampler
        batches = []
        rng.seed_rngs(0)

        # Restore everything.
        rng.load_rng_state_dict(rng_state)
        dataloader, sampler = self.prepare(data.DatasetSplit.VALID, skip_seed=True)
        assert isinstance(sampler, hlds.ResumableRandomSampler)
        sampler.set_epoch(0)
        sampler.start_iter = half_step

        # Continue.
        for batch in dataloader:
            batches.append(batch)

        self.check_batches(exp_batches[half_step:], batches)

    def test_resume_sequential(self) -> None:
        dataloader, sampler = self.prepare(data.DatasetSplit.TEST)
        assert isinstance(sampler, hlds.ResumableSequentialSampler)

        half_step = len(dataloader) // 2
        exp_batches = self.get_expected_test_batches()

        batches = []
        sampler.set_epoch(0)
        for step, batch in enumerate(dataloader):
            if step >= half_step:
                break
            batches.append(batch)

        self.check_batches(exp_batches[:half_step], batches)

        # Grab the state and clear everything.
        rng_state = rng.get_rng_state_dict()

        del dataloader, sampler
        batches = []
        rng.seed_rngs(0)

        # Restore everything.
        rng.load_rng_state_dict(rng_state)
        dataloader, sampler = self.prepare(data.DatasetSplit.TEST, skip_seed=True)
        assert isinstance(sampler, hlds.ResumableSequentialSampler)
        sampler.set_epoch(0)
        sampler.start_iter = half_step

        # Continue.
        for batch in dataloader:
            batches.append(batch)

        self.check_batches(exp_batches[half_step:], batches)

    def test_sampler_registry(self) -> None:
        registered_names = [
            "ResumableRandomSampler",
            "ResumableSequentialSampler",
            "ResumableDistributedSampler",
        ]
        assert len(typing.cast(typing.Sized, hlds.SAMPLER_REGISTRY.keys())) == len(
            registered_names
        )
        for name in registered_names:
            assert name in hlds.SAMPLER_REGISTRY

    def test_dataset_getters(self) -> None:
        datamodule = SampleDataModule()
        datamodule.setup()

        assert datamodule.train_dataset is not None
        assert datamodule.valid_dataset is not None
        assert datamodule.test_dataset is not None

        assert isinstance(datamodule.train_dataset, RandomDataset)
        assert isinstance(datamodule.valid_dataset, SequentialDataset)
        assert isinstance(datamodule.test_dataset, SequentialDataset)


if __name__ == "__main__":
    t = TestDataModule()
    t.test_resume_sequential()



import typing

from torch import nn

from helios import core

LOSS_REGISTRY = core.Registry("loss")
"""
Global instance of the registry for loss functions.

Example:
    .. code-block:: python

        import helios.losses as hll

        # This automatically registers your loss function.
        @hll.LOSS_REGISTRY.register
        class MyLoss:
            ...

        # Alternatively you can manually register a loss function like this:
        hll.LOSS_REGISTRY.register(MyLoss)
"""


def create_loss(type_name: str, *args: typing.Any, **kwargs: typing.Any) -> nn.Module:
    """
    Create the loss function for the given type.

    Args:
        type_name: the type of the loss to create.
        args: positional arguments to pass into the loss.
        kwargs: keyword arguments to pass into the loss.
    Returns:
        The loss function.
    """
    return LOSS_REGISTRY.get(type_name)(*args, **kwargs)



try:
    import optuna
except ImportError as e:
    raise ImportError("error: OptunaPlugin requires Optuna to be installed") from e
import typing

import torch

import helios.core.distributed as dist
import helios.model as hlm
import helios.plugins as hlp
import helios.trainer as hlt

# Ignore private member access.
# ruff: noqa: SLF001

_PRUNED_KEY = "ddp_hl:pruned"
_CYCLE_KEY = "ddp_hl:cycle"
_ORIG_NUMBER_KEY = "hl:id"


@hlp.PLUGIN_REGISTRY.register
class OptunaPlugin(hlp.Plugin):
    """
    Plug-in to do hyper-parameter tuning with Optuna.

    This plug-in integrates `Optuna <https://optuna.readthedocs.io/en/stable/>`__ into the
    training system in order to provide hyper-parameter tuning. The plug-in provides the
    following functionality:

    #. Automatic handling of trial pruning.
    #. Automatic reporting of metrics.
    #. Exception registration for trial pruning.
    #. Easy integration with Helios' checkpoint system to continue stopped trials.

    .. warning::
        This plug-in **requires** Optuna to be installed before being used. If it isn't,
        then :py:exc:`ImportError` is raised.

    Example:
        .. code-block:: python

            import helios.plugins as hlp
            import optuna

            def objective(trial: optuna.Trial) -> float:
                datamodule = ...
                model = ...
                plugin = hlp.optuna.OptunaPlugin(trial, "accuracy")

                trainer = ...

                # Automatically registers the plug-in with the trainer.
                plugin.configure_trainer(trainer)

                # This can be skipped if you don't want the auto-resume functionality or
                # if you wish to manage it yourself.
                plugin.configure_model(model)

                trainer.fit(model, datamodule)
                plugin.check_pruned()
                return model.metrics["accuracy"]

            def main():
                # Note that the plug-in requires the storage to be persistent.
                study = optuna.create_study(storage="sqlite:///example.db", ...)
                study.optimize(objective, ...)

    Args:
        trial: the Optuna trial.
        metric_name: the name of the metric to monitor. This assumes the name will be
            present in the :py:attr:`~helios.model.model.Model.metrics` table.
    """

    plugin_id = "optuna"

    def __init__(self, trial: optuna.Trial, metric_name: str) -> None:
        """Create the plug-in."""
        super().__init__(self.plugin_id)
        self._trial = trial
        self._metric_name = metric_name
        self._last_cycle: int = 0

        self.unique_overrides.should_training_stop = True

    @property
    def trial(self) -> optuna.Trial:
        """Return the trial."""
        return self._trial

    @trial.setter
    def trial(self, t: optuna.Trial) -> None:
        self._trial = t

    @classmethod
    def enqueue_failed_trials(cls, study: optuna.study.Study) -> None:
        """
        Enqueue any failed trials so they can be re-run.

        This will add any failed trials from a previous run. This is used for cases when
        the study had to be stopped due to an error, exception, or by the user, allowing
        trials that didn't finish to complete.

        This function works in tandem with
        :py:meth:`~helios.plugins.optuna.OptunaPlugin.configure_model` to ensure that when
        the failed trial is re-run, the original save name is restored so any saved
        checkpoints can be re-used so the trial can continue instead of starting from
        scratch.

        .. note::
            Only trials that fail but haven't been completed will be enqueued by this
            function. If a trial fails and is completed later on, it will be skipped.

        .. warning::
            Depending on the reason for a trial failing, it is possible for this function
            to re-add trials that will continue to fail. If you require special handling,
            you may override this function to achieve your desired behaviour.

        Args:
            study: the study to get the failed trials from and enqueue them.
        """
        failed_but_completed: list[optuna.trial.FrozenTrial] = []
        failed: dict[int, optuna.trial.FrozenTrial] = {}
        for trial in study.trials:
            if (
                trial.state == optuna.trial.TrialState.COMPLETE
                and _ORIG_NUMBER_KEY in trial.user_attrs
            ):
                failed_but_completed.append(trial)

            if trial.state == optuna.trial.TrialState.FAIL:
                failed[trial.number] = trial

        for trial in failed_but_completed:
            trial_num = trial.user_attrs[_ORIG_NUMBER_KEY]
            failed.pop(trial_num, None)

        for _, trial in failed.items():
            study.enqueue_trial(trial.params, {_ORIG_NUMBER_KEY: trial.number})

    def configure_trainer(self, trainer: hlt.Trainer) -> None:
        """
        Configure the trainer with the required settings.

        This will do two things:

        #. Register the plug-in itself with the trainer.
        #. Append the trial pruned exception to the trainer.

        Args:
            trainer: the trainer instance.
        """
        self._register_in_trainer(trainer)
        self._append_train_exceptions(optuna.TrialPruned, trainer)

    def configure_model(self, model: hlm.Model) -> None:
        """
        Configure the model to set the trial number into the save name.

        This will alter the :py:attr:`~helios.model.model.Model.save_name` property of the
        model by appending :code:`_trial-<trial-numer>`.

        Args:
            model: the model instance.
        """
        n_trial = self.trial.user_attrs.get(_ORIG_NUMBER_KEY, self.trial.number)
        model._save_name = model._save_name + f"_trial-{n_trial}"

    def suggest(self, type_name: str, name: str, **kwargs: typing.Any) -> typing.Any:
        """
        Generically Wrap the ``suggest_`` family of functions of the optuna trial.

        This function can be used to easily invoke the corresponding ``suggest_`` function
        from the Optuna trial held by the plug-in without having to manually type each
        individual function. This lets you write generic code that can be controlled by an
        external source (such as command line arguments or a config table). The function
        wraps the following functions:

        .. list-table:: Suggestion Functions
            :header-rows: 1

            * - Function
              - Name
            * - ``optuna.Trial.suggest_categorical``
              - categorical
            * - ``optuna.Trial.suggest_int``
              - int
            * - ``optuna.Trial.suggest_float``
              - float

        .. warning::
            Functions that are marked as deprecated by Optuna are *not* included in this
            wrapper.

        .. note::
            You can find the exact arguments for each function `here
            <https://optuna.readthedocs.io/en/stable/reference/generated/optuna.trial.Trial.html>`__.

        Example:
            .. code-block:: python

                import helios.plugin as hlp
                import optuna

                def objective(trial: optuna.Trial) -> float:
                    plugin = hlp.optuna.OptunaPlugin(trial, "accuracy")
                    # ... configure model and trainer.

                    val1 = plugin.suggest("categorical", "val1", choices=[1, 2, 3])
                    val2 = plugin.suggest("int", "val2", low=0, high=10)
                    val3 = plugin.suggest("float", "val3", low=0, high=1)

        Args:
            type_name: the name of the type to suggest from.
            name: a parameter name
            **kwargs: keyword arguments to the corresponding suggest function.

        Raises:
            KeyError: if the value passed in to ``type_name`` is not recognised.
        """
        if type_name not in ("categorical", "float", "int"):
            raise KeyError(f"error: {type_name} is not a valid suggestion type.")

        fn = getattr(self._trial, f"suggest_{type_name}")
        return fn(name, **kwargs)

    def setup(self) -> None:
        """
        Configure the plug-in.

        Raises:
            ValueError: if the study wasn't created with persistent storage.
        """
        if self.is_distributed and not (
            isinstance(self.trial.study._storage, optuna.storages._CachedStorage)
            and isinstance(self.trial.study._storage._backend, optuna.storages.RDBStorage)
        ):
            raise ValueError(
                "error: optuna integration supports only optuna.storages.RDBStorage "
                "in distributed mode"
            )

    def on_validation_end(self, validation_cycle: int) -> None:
        """
        Report metrics to the trial.

        .. note::
            In distributed training, only rank 0 will report the metrics to the trial.

        Args:
            validation_cycle: the current validation cycle.
        """
        model = self.trainer.model
        if not model.metrics or self._metric_name not in model.metrics:
            return

        if self.rank == 0:
            self.trial.report(model.metrics[self._metric_name], validation_cycle)
        self._last_cycle = validation_cycle

    def should_training_stop(self) -> bool:
        """
        Handle trial pruning.

        Returns:
            True if the trial should be pruned, false otherwise.
        """
        should_stop = False
        if self.rank == 0:
            should_stop = self.trial.should_prune()

        # Sync the value across all processes (if using distributed training).
        if self.is_distributed:
            t = dist.all_reduce_tensors(torch.tensor(should_stop).to(self.device))
            should_stop = t.item()  # type: ignore[assignment]

        if should_stop and self.rank == 0:
            self.trial.set_user_attr(_PRUNED_KEY, True)
            self.trial.set_user_attr(_CYCLE_KEY, self._last_cycle)

        return should_stop

    def on_training_end(self) -> None:
        """
        Clean-up on training end.

        If training is non-distributed and the trial was pruned, then this function will
        do the following:

        #. Call :py:meth:`~helios.model.model.Model.on_training_end` to ensure metrics are
           correctly logged (if using).
        #. Raise :py:exc:`optuna.TrialPruned` exception to signal the trial was pruned.

        If training is distributed, this function does nothing.

        Raises:
            TrialPruned: if the trial was pruned.
        """
        if not self.is_distributed and self.trial.should_prune():
            self.trainer.model.on_training_end()
            raise optuna.TrialPruned(f"Pruned on validation cycle {self._last_cycle}")

    def check_pruned(self) -> None:
        """
        Ensure pruned distributed trials are correctly handled.

        Due to the way distributed training works, we can't raise an exception within the
        distributed processes, so we have to do it after we return to the main process.
        If the trial was pruned, this function will raise :py:exc:`optuna.TrialPruned`. If
        distributed training wasn't used, this function does nothing.

        .. warning::
            You *must* ensure this function is called after
            :py:meth:`~helios.trainer.Trainer.fit` to ensure pruning works correctly.

        Raises:
            TrialPruned: if the trial was pruned.
        """
        trial_id = self.trial._trial_id
        study = self.trial.study
        trial = study._storage._backend.get_trial(trial_id)  # type: ignore[attr-defined]
        is_pruned = trial.user_attrs.get(_PRUNED_KEY)
        val_cycle = trial.user_attrs.get(_CYCLE_KEY)

        if is_pruned is None or val_cycle is None:
            return

        if is_pruned:
            raise optuna.TrialPruned(f"Pruned on validation cycle {val_cycle}")

    def state_dict(self) -> dict[str, typing.Any]:
        """
        Get the state of the current trial.

        This will return the parameters to be optimised for the current trial.

        Returns:
            The parameters of the trial.
        """
        return self._trial.params
